<?php
/**
 * Interaction witch Woocommerce.
 *
 * @package woocommerce-payment-rukassa
 */

namespace RuKassa\Payment;

defined( 'ABSPATH' ) || exit;

use WC_HTTPS;
use WC_Order;
use WC_Payment_Gateway;
use WP_REST_Server;
use WP_Error;
use Exception;
use ErrorException;

/**
 * RuKassa payment gateway.
 *
 * @package RuKassa\Payment
 */
class Gateway extends WC_Payment_Gateway {
	/**
	 * ID of the class extending the settings API. Used in option names.
	 *
	 * @var string
	 */
	public $id = 'rukassa';

	/**
	 * The notification URL.
	 *
	 * @var string
	 */
	protected $success_url;

	/**
	 * The notification URL.
	 *
	 * @var string
	 */
	protected $fail_url;

	/**
	 * The notification URL.
	 *
	 * @var string
	 */
	protected $notification_url;

	/**
	 * The shop id.
	 *
	 * @var string
	 */
	protected $shop_id;

	/**
	 * The shop token.
	 *
	 * @var string
	 */
	protected $shop_token;

	/**
	 * The supported ways.
	 *
	 * @var string
	 */
	protected $method_supports;

	/**
	 * The title icon.
	 *
	 * @var string
	 */
	protected $title_icon;

	/**
	 * The use HTML flag.
	 *
	 * @var bool
	 */
	protected $use_html;

	/**
	 * The use debug flag.
	 *
	 * @var bool
	 */
	protected $use_debug;

	/**
	 * The title icon HTML.
	 *
	 * @var string
	 */
	protected $title_icon_html;

	/**
	 * The icon HTML.
	 *
	 * @var string
	 */
	protected $icon_html;

	/**
	 * The method description HTML.
	 *
	 * @var string
	 */
	protected $method_description_html;

	/**
	 * Woocommerce logger.
	 *
	 * @var /WC_Logger_Interface
	 */
	protected $logger;

	/**
	 * Gateway constructor.
	 */
	public function __construct() {
		/*
		 * translators:
		 * ru_RU: RuKassa
		 */
		$this->method_title = __( 'RuKassa Pay', 'woocommerce_payment_rukassa' );

		/*
		 * translators:
		 * ru_RU: Оплата через: VISA, MasterCard, МИР, Другое
		 */
		$this->method_supports = __( 'Payment over: VISA, MasterCard, MIR, more+', 'woocommerce_payment_rukassa' );

		/*
		 * translators:
		 * ru_RU: RuKassa оплата
		 */
		$this->method_description = __( 'RuKassa Pay', 'woocommerce_payment_rukassa' );

		// Getaway accept billing and refunding.
		$this->supports = [ 'products', 'refunds' ];

		// Setup logger.
		$this->logger = wc_get_logger();

		// Setup icons.
		$this->icon       = plugins_url( 'assets/payments.svg', __DIR__ );
		$this->title_icon = plugins_url( 'assets/logo_kassa.svg', __DIR__ );

		// Setup readonly props.
		$this->success_url = site_url() . '/?wc-api=' . $this->id . '&rukassa=success';
		$this->fail_url = site_url() . '/?wc-api=' . $this->id . '&rukassa=fail';
		$this->notification_url = site_url() . '/?wc-api=' . $this->id;

		// Initialise config and form.
		$this->init_form_fields();
		$this->init_settings();
		$this->init_options();

		// Set up HTML fragments.
		$this->title_icon_html         = '<img src="' . WC_HTTPS::force_https_url( $this->title_icon ) . '" alt="' . esc_attr( $this->title ) . '" class="rukassa" />';
		$this->icon_html               = '<img src="' . WC_HTTPS::force_https_url( $this->icon ) . '" alt="' . esc_attr( $this->method_supports ) . '" />';
		$this->method_description_html = $this->method_description . PHP_EOL .

			/*
			 * translators:
			 * ru_RU: Для начала работы с сервисом RuKassa необходима <a href="https://lk.rukassa.is/" target="_blank">регистрация магазина</a>.
			 */
			__( 'To start working with the RuKassa cash service, you need to <a href="https://lk.rukassa.is/" target="_blank">register a store</a>.', 'woocommerce_payment_rukassa' ) . PHP_EOL .

			/*
			 * translators:
			 * ru_RU: Так же, для вас доступен <a href="https://lk.rukassa.is/" target="_blank">демонстрационный стенд</a>.
			 */
			__( 'Also, a <a href="https://lk.rukassa.is/" target="_blank">demonstration stand</a> is available for you.', 'woocommerce_payment_rukassa' );

		// Setup CURL options.
		$options = [];
		$ca_path = dirname( __DIR__ ) . DIRECTORY_SEPARATOR . 'cacert.pem';
		if ( is_file( $ca_path ) ) {
			$options[ CURLOPT_SSL_VERIFYPEER ] = true;
			$options[ CURLOPT_SSL_VERIFYHOST ] = 2;
			$options[ CURLOPT_CAINFO         ] = $ca_path;
		}

		// Capture API callback.
		add_action( "woocommerce_api_{$this->id}", [ $this, 'woocommerce_api' ] );

		// Prevent html escape.
		add_filter( 'esc_html', [ $this, 'title_esc_html' ], 50, 2 );

		if ( is_admin() ) {
			// Capture options change.
			add_action( "woocommerce_update_options_payment_gateways_{$this->id}", [ $this, 'process_admin_options' ] );
		} else {

			// Add frontend style.
			wp_enqueue_style(
				'woocommerce-payment-rukassa',
				plugins_url( '/assets/rukassa.css', __DIR__ ),
				[],
				filemtime( dirname( __DIR__ ) . DIRECTORY_SEPARATOR . 'assets' . DIRECTORY_SEPARATOR . 'rukassa.css' )
			);
		}
	}

	/**
	 * Log debug message.
	 *
	 * @param string $message The message.
	 * @param array  $context The context.
	 */
	protected function log( $message, $context ) {
		if ( $this->use_debug ) {
			//phpcs:disable WordPress.PHP.DevelopmentFunctions.error_log_print_r -- Because on debug mode only.
			$this->logger->debug( $message . ': ' . print_r( $context, true ) );
			//phpcs:enable WordPress.PHP.DevelopmentFunctions.error_log_print_r
		}
	}

	/**
	 * Load options.
	 */
	protected function init_options() {
		$this->title       = $this->get_option( 'title', $this->method_title );
		$this->use_html    = $this->get_option( 'use_html', 'yes' ) === 'yes';
		$this->description = $this->get_option( 'description', $this->method_description );
		$this->shop_id     = $this->get_option( 'shop_id' );
		$this->shop_token  = $this->get_option( 'shop_token' );
		$this->use_debug   = $this->get_option( 'use_debug' ) === 'yes';
	}

	/**
	 * Process settings change.
	 *
	 * @return bool
	 */
	public function process_admin_options() {
		/**
		 * The WP DB instance.
		 *
		 * @var \wpdb $wpdb
		 */
		global $wpdb;

		$result = parent::process_admin_options();
		$this->init_options();

		/**
		 * Query for update payment title on old bills.
		 *
		 * @noinspection SqlResolve
		 */
		$wpdb->query( $wpdb->prepare(
			"UPDATE $wpdb->postmeta AS t1 LEFT JOIN $wpdb->postmeta AS t2 ON t1.`post_id` = t2.`post_id` SET t1.`meta_value` = %s WHERE t1.`meta_key` = '_payment_method_title' AND t2.`meta_key` = '_payment_method' AND t2.`meta_value` = %s",
			$this->get_title(),
			$this->id
		) );

		return $result;
	}

	/**
	 * Get method description.
	 *
	 * @return string
	 */
	public function get_method_description() {
		return apply_filters( 'woocommerce_gateway_method_description', $this->method_description_html, $this );
	}

	/**
	 * Return the gateway's title.
	 *
	 * @return string
	 */
	public function get_title() {
		return apply_filters( 'woocommerce_gateway_title', $this->use_html ? $this->title_icon_html : $this->title, $this->id );
	}

	/**
	 * Return title unescaped always.
	 *
	 * @param string $safe_text The safe text.
	 * @param string $text The original text.
	 *
	 * @return mixed
	 */
	public function title_esc_html( $safe_text, $text ) {
		return $text === $this->title_icon_html ? $text : $safe_text;
	}

	/**
	 * Return the gateway's icon.
	 *
	 * @return string
	 */
	public function get_icon() {
		return apply_filters( 'woocommerce_gateway_icon', $this->icon_html, $this->id );
	}

	/**
	 * Check gateway not ready for use.
	 *
	 * @return bool
	 */
	public function needs_setup() {
		return ! empty( $this->shop_id ) || ! empty( $this->shop_token );
	}

	/**
	 * Initialise settings.
	 */
	public function init_settings() {
		parent::init_settings();

		// Readonly option.
		$this->settings['success_url'] = $this->success_url;
		$this->settings['fail_url'] = $this->fail_url;
		$this->settings['notification_url'] = $this->notification_url;
	}

	/**
	 * Set up config page fields.
	 */
	public function init_form_fields() {
		$this->form_fields = [
			'enabled'          => [

				/*
				 * translators:
				 * ru_RU: Включение / выключение
				 */
				'title'   => __( 'Turn On/off', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Активировать модуль оплаты %1$s
				 * 1: Название модуля
				 */
				'label'   => sprintf( __( 'Activate payment module %1$s', 'woocommerce_payment_rukassa' ), $this->method_title ),
				'type'    => 'checkbox',
				'default' => 'no',
			],
			'title'            => [

				/*
				 * translators:
				 * ru_RU: Заголовок
				 */
				'title'       => __( 'Title', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Название метода оплаты, отображаемое клиентам.
				 */
				'description' => __( 'The name of the payment method displayed to customers.', 'woocommerce_payment_rukassa' ),
				'type'        => 'text',
				'default'     => $this->method_title,
			],
			'use_html'         => [

				/*
				 * translators:
				 * ru_RU: Использовать HTML
				 */
				'title'   => __( 'Use logo', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Модуль оплаты %1$s будет отображать логотип вместо названия
				 * 1: Название модуля
				 */
				'label'   => sprintf( __( 'Payment module %1$s will show logo instead name', 'woocommerce_payment_rukassa' ), $this->method_title ),
				'type'    => 'checkbox',
				'default' => 'yes',
			],
			'description'      => [

				/*
				 * translators:
				 * ru_RU: Описание
				 */
				'title'       => __( 'Description', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Описание метода оплаты, отображаемое клиентам.
				 */
				'description' => __( 'Description of the payment method displayed to customers.', 'woocommerce_payment_rukassa' ),
				'type'        => 'textarea',
				'default'     => $this->method_description,
			],
			'shop_id'       => [

				/*
				 * translators:
				 * ru_RU: ID магазина
				 */
				'title'       => __( 'Shop ID', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: ID вашего магазина.
				 */
				'description' => __( 'The ID your merchant.', 'woocommerce_payment_rukassa' ),
				'type'        => 'text',
			],
			'shop_token'       => [

				/*
				 * translators:
				 * ru_RU: Токен магазина
				 */
				'title'       => __( 'Shop token', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Токен вашего магазина.
				 */
				'description' => __( 'The token your merchant.', 'woocommerce_payment_rukassa' ),
				'type'        => 'password',
			],
			'success_url' => [

				/*
				 * translators:
				 * ru_RU: Адрес для уведомлений
				 */
				'title'       => __( 'Success address', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Установите это значение в настройках магазина платежной системы.
				 */
				'description' => __( 'Set this value in the payment system store settings.', 'woocommerce_payment_rukassa' ),
				'type'        => 'text',
				'disabled'    => true,
				'default'     => $this->success_url,
			],
			'fail_url' => [

				/*
				 * translators:
				 * ru_RU: Адрес для уведомлений
				 */
				'title'       => __( 'Fail address', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Установите это значение в настройках магазина платежной системы.
				 */
				'description' => __( 'Set this value in the payment system store settings.', 'woocommerce_payment_rukassa' ),
				'type'        => 'text',
				'disabled'    => true,
				'default'     => $this->fail_url,
			],
			'notification_url' => [

				/*
				 * translators:
				 * ru_RU: Адрес для уведомлений
				 */
				'title'       => __( 'Notification address', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Установите это значение в настройках магазина платежной системы.
				 */
				'description' => __( 'Set this value in the payment system store settings.', 'woocommerce_payment_rukassa' ),
				'type'        => 'text',
				'disabled'    => true,
				'default'     => $this->notification_url,
			],
			'use_debug'        => [

				/*
				 * translators:
				 * ru_RU: Режим отладки
				 */
				'title'       => __( 'Debug mode', 'woocommerce_payment_rukassa' ),

				/*
				 * translators:
				 * ru_RU: Включить логирование запросов API RuKassa.
				 */
				'description' => __( 'Enable logging RuKassa API requests.', 'woocommerce_payment_rukassa' ),
				'type'        => 'checkbox',
				'default'     => 'not',
			],
		];
	}
	
	/**
	 * Process RuKassa success.
	 */
    public function successPayment(string $url): void
    {
        WC()->cart->empty_cart();
        wp_redirect($url);
    }

    /**
	 * Process RuKassa fail.
	 */
    public function failPayment(string $url): void
    {
        wp_redirect($url);
    }

	/**
	 * Process RuKassa notification.
	 */
	public function woocommerce_api() {
		
		if ($_REQUEST['rukassa'] === 'success') {
			// Get order.
			$orders = wc_get_orders( [
				'limit'          => 1,
				'transaction_id' => $notice['order_id'],
			] );
			$order = reset( $orders );
			$this->successPayment($this->get_return_url($order));
			exit;
        }
		
		if ($_REQUEST['rukassa'] === 'fail') {
			// Get order.
			$orders = wc_get_orders( [
				'limit'          => 1,
				'transaction_id' => $notice['order_id'],
			] );
			$order = reset( $orders );
			$this->failPayment($this->get_return_url($order));
			exit;
        }
		
		// Get request data.
		$sign   = array_key_exists( 'HTTP_SIGNATURE', $_SERVER ) ? wp_unslash( $_SERVER['HTTP_SIGNATURE'] ) : '';  // phpcs:ignore WordPress.VIP
		$body   = WP_REST_Server::get_raw_data();
		//$notice = json_decode( $body, true );
		$notice = $_POST;
		
		// Check signature.
		//$result = $this->bill_payments->checkNotificationSignature( $sign, $notice, $this->secret_key );
		$sign2 = hash_hmac('sha256', $notice['id'] . '|' . $notice['createdDateTime'] . '|' . $notice['amount'], $this->shop_token);
		$result = ($sign == $sign2) ? true : false;
		if($result) {
			$result = ($notice['in_amount'] >= $notice['amount']) ? true : false;
		}
		
		$this->log(
			$result ?

				/*
				 * translators:
				 * ru_RU: Получено действительное уведомление
				 */
				__( 'Received valid notification', 'woocommerce_payment_rukassa' ) :

				/*
				 * translators:
				 * ru_RU: Получено недействительное уведомление
				 */
				__( 'Received invalid notification', 'woocommerce_payment_rukassa' ),
			[
				'sign'   => $sign,
				'notice' => $notice,
			]
		);
		if ( ! $result ) {
			wp_send_json( [ 'error' => 403 ], 403 );
		}

		// Get order.
		$orders = wc_get_orders( [
			'limit'          => 1,
			'transaction_id' => $notice['order_id'],
		] );
		/**
		 * The order.
		 *
		 * @var WC_Order $order
		 */
		$order = reset( $orders );

		// Process status.
		switch ( $notice['status'] ) {
			case 'WAIT':
				$order->update_status( 'pending' );
				break;
			case 'PAID':
				$order->payment_complete();
				break;
			case 'CANCEL':
				$order->update_status( 'canceled' );
				break;
			case 'REFAUND':
				$order->update_status( 'refunded' );
				break;
		}
		wp_send_json( [ 'error' => 0 ], 200 );
	}

	/**
	 * Process the payment and return the result.
	 *
	 * @param int $order_id The order ID.
	 * @return array
	 */
	public function process_payment( $order_id ) {
		// Get processed data.
		$order   = wc_get_order( $order_id );
		$bill_id = $order->get_transaction_id();

		// Return notice on trow.
		try {
			// Need to create bill transaction.
			if ( $order->get_status() === 'pending' ) {

				$bill_id = time() . mt_rand();
				
				$params = [
					'shop_id'		=> $this->shop_id,
					'order_id'		=> $bill_id,
					'amount'		=> $order->get_total(),
					'token'			=> $this->shop_token,
					'data'			=> json_encode([
						'phone' 	=> $order->get_billing_phone(),
						'email' 	=> $order->get_billing_email(),
						'account'   => $order->get_user_id(),
					]),
				];
				
				$ch = curl_init('https://lk.rukassa.is/api/v1/create');
				curl_setopt($ch, CURLOPT_POST, 1);
				curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($params, '', '&'));
				curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
				curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
				curl_setopt($ch, CURLOPT_HEADER, false);
				$bill = json_decode(curl_exec($ch));
				//$bill = curl_exec($ch);
				curl_close($ch);
				
				$this->log(

					/*
					 * translators:
					 * ru_RU: Создан счет
					 */
					__( 'Create bill', 'woocommerce_payment_rukassa' ),
					[
						'params' => $params,
						'bill'   => $bill,
					]
				);
				$order->set_transaction_id( $bill_id );
				$order->save();

				// Reduce stock levels.
				wc_reduce_stock_levels( $order->get_id() );

				// Remove cart.
				wc_empty_cart();
			} elseif ( ! $order->is_paid() && $order->get_status() === 'cancelled' ) {
				//$bill = $this->bill_payments->cancelBill( $bill_id );
				$this->log(

					/*
					 * translators:
					 * ru_RU: Завершен счет
					 */
					__( 'Cancel bill', 'woocommerce_payment_rukassa' ),
					[
						'bill_id' => $bill_id,
						'bill'    => '',
					]
				);
			} else {
				//$bill = $this->bill_payments->getBillInfo( $bill_id );
				$this->log(

					/*
					 * translators:
					 * ru_RU: Получена информация о счете
					 */
					__( 'Get bill info', 'woocommerce_payment_rukassa' ),
					[
						'bill_id' => $bill_id,
						'bill'    => '',
					]
				);
			}
		} catch ( Exception $exception ) {
			/*
			 * translators:
			 * ru_RU: Ошибка обращения к RuKassa
			 */
			$message = __( 'RuKassa request error', 'woocommerce_payment_rukassa' );
			wc_add_wp_error_notices( new WP_Error(
				$exception->getCode(),
				$message . '<br>' . $exception->getMessage(),
				$exception
			) );
			return [ 'result' => 'fail' ];
		}

		// Return thank you redirect.
		$result = [
			'result'   => 'success',
			'success'  => $this->get_return_url( $order ),
			'redirect' => $bill->url,
		];

		// Detect AJAX.
		$request = array_key_exists( 'HTTP_X_REQUESTED_WITH', $_SERVER ) ? wp_unslash( $_SERVER['HTTP_X_REQUESTED_WITH'] ) : ''; // phpcs:ignore WordPress.VIP
		if ( strtolower( $request ) === 'xmlhttprequest' ) {
			wp_send_json( apply_filters( 'woocommerce_payment_successful_result', $result, $order_id ) );
		}

		return $result;
	}

	/**
	 * Process refunding.
	 *
	 * @param int    $order_id The order ID.
	 * @param null   $amount   The refund amount.
	 * @param string $reason   The refund reason.
	 * @return bool|WP_Error
	 */
	public function process_refund( $order_id, $amount = null, $reason = null ) {
		// Extract data.
		$order   = wc_get_order( $order_id );
		$bill_id = $order->get_transaction_id();

		// Generated refund transaction ID.
		try {
			$refund_id = $this->bill_payments->generateId();
		} catch ( Exception $exception ) {
			return new WP_Error(
				$exception->getCode(),
				$exception->getMessage(),
				$exception
			);
		}

		// Refund transaction.
		try {
			$refund = $this->bill_payments->refund( $bill_id, $refund_id, $amount, $order->get_currency() );
			$this->log(

				/*
				 * translators:
				 * ru_RU: Создан возврат по счету
				 */
				__( 'Create bill refund', 'woocommerce_payment_rukassa' ),
				[
					'bill_id'   => $bill_id,
					'refund_id' => $refund_id,
					'amount'    => $amount,
					'currency'  => $order->get_currency(),
					'refund'    => $refund,
				]
			);
		} catch ( Exception $exception ) {
			return new WP_Error(
				$exception->getCode(),
				$exception->getMessage(),
				$exception
			);
		}

		// Process result.
		switch ( $refund['status'] ) {
			case 'PARTIAL':
			case 'FULL':
				$order->add_order_note( sprintf(

					/*
					 * translators:
					 * ru_RU: Выполнен возврат %1$s
					 * 1: ID транзакции
					 */
					__( 'Completed refund  %1$s', 'woocommerce_payment_rukassa' ),
					$refund_id
				) );
				return true;
		}
		return false;
	}
}
